import unittest
from datetime import date
from decimal import Decimal

from accounts.metadata import RateType


class TestRateType(unittest.TestCase):
    def setUp(self):
        self.value_date = date(2019, 1, 1)
        self.payment_rates = RateType(name="payments", label="Wire Payment Tier Rates")
        self.payment_rates.add_tier(self.value_date, Decimal(10), Decimal(0))
        self.payment_rates.add_tier(self.value_date, Decimal(100), Decimal(5))
        self.payment_rates.add_tier(self.value_date, Decimal(1000), Decimal(4))
        self.payment_rates.add_tier(self.value_date, Decimal(1000000), Decimal(3))

    def test_get_rate(self):
        payment_rates = RateType(name="payments", label="Payment Rates")

        payment_rates.add_tier(self.value_date, Decimal(1000), Decimal(0))
        payment_rates.add_tier(self.value_date, Decimal(10000), Decimal(0.01))

        self.assertEqual(payment_rates.get_rate(self.value_date, Decimal(0.000000001)), Decimal(0))

    def test_get_fee_success(self):
        self.assertEqual(self.payment_rates.get_fee(self.value_date, Decimal(0), Decimal(5)), Decimal(0))
        self.assertEqual(self.payment_rates.get_fee(self.value_date, Decimal(5), Decimal(15)), Decimal(25))
        self.assertEqual(self.payment_rates.get_fee(self.value_date, Decimal(15), Decimal(55)), Decimal(200))
        self.assertEqual(self.payment_rates.get_fee(self.value_date, Decimal(55), Decimal(1005)), Decimal(3840))

    def test_get_daily_fee(self):
        rt_users = RateType(name="userRates", label="test")
        rt_users.add_tier(self.value_date, Decimal(3), Decimal(30))
        rt_users.add_tier(self.value_date, Decimal(10), Decimal(25))
        rt_users.add_tier(self.value_date, Decimal(1000), Decimal(10))

        value_date = date(2020, 4, 20)

        self.assertEqual(rt_users.get_daily_fee(Decimal(1), value_date), Decimal(1))
        self.assertEqual(rt_users.get_daily_fee(Decimal(5), value_date), Decimal(3 * 30 + 2 * 25) / (Decimal(30)))
        self.assertEqual(rt_users.get_daily_fee(Decimal(12), value_date),
                         Decimal(3 * 30 + 7 * 25 + 2 * 10) / Decimal(30))


if __name__ == '__main__':
    unittest.main()



from calendar import monthrange
from datetime import date
from decimal import Decimal
from enum import Enum
from typing import List, Optional, Dict
from pydantic import BaseModel

from accounts.utility import CustomEncoder


class TransactionOperation(Enum):
    CREDIT = 'credit'
    DEBIT = 'debit'
    SET = 'set'


class DataType(Enum):
    DECIMAL = 'decimal'
    STRING = 'string'
    BOOLEAN = 'boolean'


class ScheduleEndType(Enum):
    NO_END = "no_end"
    END_REPEATS = "end_repeats"
    END_DATE = "end_date"


class BusinessDayAdjustment(Enum):
    NO_ADJUSTMENT = "no_adjustment"
    NEXT_WORKING_DAY = "next_working_day"
    PREVIOUS_WORKING_DAY = "previous_working_day"
    CLOSEST_WORKING_DAY = "closest_working_day"


class ScheduleFrequency(Enum):
    DAILY = "daily"
    MONTHLY = "monthly"


class ScheduledTransactionTiming(Enum):
    START_OF_DAY = "start_of_day"
    END_OF_DAY = "end_of_day"


class PositionType(BaseModel):
    name: str
    label: str


class DateType(BaseModel):
    name: str
    label: str
    required: bool = True


class RateTier(BaseModel):
    from_amount: Decimal
    to_amount: Decimal
    rate: Decimal


class RateType(BaseModel):
    name: str
    label: str
    rate_tiers: Dict[str, List[RateTier]] = {}

    class Config:
        json_encoders = {Dict: CustomEncoder()}

    @staticmethod
    def __get_key(value_date):
        return value_date.strftime("%Y-%m-%d")

    def add_tier(self, value_date: date, to_amount: Decimal, rate: Decimal):
        key = self.__get_key(value_date)
        if key not in self.rate_tiers:
            self.rate_tiers[key] = []

        rate_tier = RateTier(from_amount=self.get_max_to_amount(value_date), to_amount=to_amount, rate=rate)
        self.rate_tiers[key].append(rate_tier)

    def get_max_to_amount(self, value_date: date):
        key = self.__get_key(value_date)
        if key not in self.rate_tiers:
            return Decimal(0)

        rate_tiers = self.rate_tiers[key]

        if len(rate_tiers) == 0:
            return Decimal(0)

        max_value = max(rate_tiers, key=lambda tier: tier.to_amount)
        return max_value.to_amount

    def __get_tiers(self, value_date) -> List[RateTier]:
        # find first date that is less than or equal to value_date
        key = self.__get_key(value_date)
        dates = [d for d in self.rate_tiers.keys() if d <= key]
        if len(dates) == 0:
            raise Exception(f"No rate tiers found for date {str(value_date)} in rate table {self.name}")
        index_date = max(dates)
        return self.rate_tiers[index_date]

    def get_rate(self, value_date: date, amount: Decimal) -> Decimal:
        applicable_tiers = (rt for rt in self.__get_tiers(value_date) if rt.from_amount <= amount <= rt.to_amount)

        # if no tiers are applicable, raise an exception
        rate_tier = next(applicable_tiers, None)
        if rate_tier is None:
            if amount < Decimal(0):
                return Decimal(0)
            else:
                raise Exception(
                    f"No rate tiers found for amount {str(amount)} on date {str(value_date)} in rate table {self.name}")

        return rate_tier.rate

    def get_daily_fee(self, users: Decimal, value_date: date):
        _, days_in_month = monthrange(value_date.year, value_date.month)
        monthly_fee = Decimal(self.get_fee(value_date, Decimal(0), users))
        return monthly_fee / Decimal(days_in_month)

    def get_fee(self, value_date: date, from_amount: Decimal, to_amount: Decimal):
        processed = from_amount
        fee = Decimal(0)

        exit_loop = False

        for rate_tier in self.__get_tiers(value_date):
            if rate_tier.from_amount <= processed < rate_tier.to_amount:
                part_processed = rate_tier.to_amount - processed

                if to_amount < rate_tier.to_amount:
                    part_processed = to_amount - processed
                    exit_loop = True

                fee = fee + part_processed * rate_tier.rate
                processed = processed + part_processed

                if exit_loop:
                    break

        return fee


class PositionRule(BaseModel):
    operation: TransactionOperation
    position_type_name: str


class TransactionType(BaseModel):
    name: str
    label: str
    position_rules: List[PositionRule] = []
    maximum_precision: bool = False

    def add_position_rule(self, transaction_operation: TransactionOperation, position_type: PositionType):
        self.position_rules.append(PositionRule(operation=transaction_operation, position_type_name=position_type.name))
        return self


class ScheduleType(BaseModel):
    name: str
    label: str
    frequency: ScheduleFrequency
    end_type: ScheduleEndType
    business_day_adjustment: BusinessDayAdjustment
    interval_expression: str
    start_date_expression: str
    end_date_expression: str = None
    number_of_repeats_expression: str = None
    include_dates_expression: str = None
    exclude_dates_expression: str = None
    editable: bool = True


class ScheduledTransaction(BaseModel):
    schedule_name: str
    timing: ScheduledTransactionTiming
    generated_transaction_type: str
    amount_expression: str


class TriggeredTransaction(BaseModel):
    trigger_transaction_type_name: str
    generated_transaction_type: str
    amount_expression: str


class PropertyType(BaseModel):
    name: str
    label: str
    data_type: DataType
    required: bool = True
    value_dated: bool = False


class InstalmentType(BaseModel):
    name: str
    label: str
    timing: ScheduledTransactionTiming
    schedule_name: str
    transaction_type: str
    property_name: str
    solve_for_zero_position: str
    solve_for_date: str


class AccountType(BaseModel):
    name: str
    label: str
    transaction_types: List[TransactionType] = []
    position_types: List[PositionType] = []
    date_types: List[DateType] = []
    rate_types: Dict[str, RateType] = {}
    triggered_transactions: List[TriggeredTransaction] = []
    schedule_types: List[ScheduleType] = []
    property_types: List[PropertyType] = []
    scheduled_transactions: List[ScheduledTransaction] = []
    instalment_type: InstalmentType = None

    def add_property_type(self, name: str, label: str, data_type: DataType, required: bool = True) -> PropertyType:
        property_type = PropertyType(name=name, label=label, data_type=data_type.value, required=required)
        self.property_types.append(property_type)
        return property_type

    def add_date_type(self, name: str, label: str) -> DateType:
        date_type = DateType(name=name, label=label)
        self.date_types.append(date_type)
        return date_type

    def add_transaction_type(self, name: str, label: str, maximum_precision: bool = False) -> TransactionType:
        transaction_type = TransactionType(name=name, label=label, maximum_precision=maximum_precision)
        self.transaction_types.append(transaction_type)
        return transaction_type

    def add_position_type(self, name: str, label: str) -> PositionType:
        position_type = PositionType(name=name, label=label)
        self.position_types.append(position_type)
        return position_type

    def add_rate_type(self, name: str, label: str) -> RateType:
        rate_type = RateType(name=name, label=label)
        self.rate_types[name] = rate_type
        return rate_type

    def add_trigger_transaction(self, trigger_transaction_type: TransactionType,
                                generated_transaction_type: TransactionType, amount_expression: str):
        trigger_transaction = TriggeredTransaction(trigger_transaction_type_name=trigger_transaction_type.name,
                                                   generated_transaction_type=generated_transaction_type.name,
                                                   amount_expression=amount_expression)
        self.triggered_transactions.append(trigger_transaction)

    def add_schedule_type(self, schedule_type: ScheduleType):
        self.schedule_types.append(schedule_type)

    def add_scheduled_transaction(self, schedule_type: ScheduleType, timing: ScheduledTransactionTiming,
                                  generated_transaction_type: TransactionType, amount_expression: str):
        scheduled_transaction = ScheduledTransaction(schedule_name=schedule_type.name, timing=timing,
                                                     generated_transaction_type=generated_transaction_type.name,
                                                     amount_expression=amount_expression)
        self.scheduled_transactions.append(scheduled_transaction)

    def get_transaction_type(self, transaction_type_name: str) -> TransactionType:
        return next(tt for tt in self.transaction_types if tt.name == transaction_type_name)

    def get_rate_type(self, rate_type_name: str):
        return next(rt for rt in self.rate_types if rt.name == rate_type_name)

    def get_trigger_transaction(self, trigger_transaction_type_name: str) -> Optional[TriggeredTransaction]:
        return next((tt for tt in self.triggered_transactions if
                     tt.trigger_transaction_type_name == trigger_transaction_type_name), None)

    def add_instalment_type(self, name: str, label: str, timing: ScheduledTransactionTiming,
                            transaction_type: str, property_name: str,
                            solve_for_zero_position: str, solve_for_date: str,
                            schedule_name: str) -> InstalmentType:

        instalment_type = InstalmentType(name=name, label=label, timing=timing,
                                         transaction_type=transaction_type,
                                         property_name=property_name,
                                         schedule_name=schedule_name,
                                         solve_for_zero_position=solve_for_zero_position,
                                         solve_for_date=solve_for_date)
        self.instalment_type = instalment_type

    def __getattr__(self, method_name):
        if method_name in self.rate_types:
            return self.rate_types[method_name]
        else:
            raise AttributeError(f'No such attribute: {method_name}')



from datetime import timedelta
from itertools import groupby
from typing import Mapping, Any
from dateutil.relativedelta import *
from pydantic import Field

from accounts.metadata import *
import scipy.optimize


class Position(BaseModel):
    amount: Decimal = Decimal(0)

    def apply_operation(self, operation: TransactionOperation, amount: Decimal):
        if operation == TransactionOperation.CREDIT:
            self.amount = self.amount + amount
        elif operation == TransactionOperation.DEBIT:
            self.amount = self.amount - amount
        else:  # Replace
            self.amount = amount

    def copy(self):
        position = Position()
        position.amount = self.amount
        return position


class Transaction(BaseModel):
    amount = Decimal(0)
    action_date: date
    value_date: date
    transaction_type: str
    amount: Decimal
    system_generated: bool

    def to_external_transaction(self):
        return ExternalTransaction(
            transaction_type_name=self.transaction_type,
            amount=self.amount,
            value_date=self.value_date
        )


class Schedule(BaseModel):
    start_date: date
    end_type: ScheduleEndType
    frequency: ScheduleFrequency
    interval: int = 0,
    adjustment: BusinessDayAdjustment = BusinessDayAdjustment.NO_ADJUSTMENT
    end_date: Optional[date]
    number_of_repeats: int = 0
    include_dates: list[date] = []
    exclude_dates: list[date] = []
    cached_dates: dict[date, date] = Field(default_factory=dict, exclude=True)

    class Config:
        # exclude the "cached_dates" field from JSON serialization
        exclude = {"cached_dates"}

    def __is_simple_daily_schedule(self):
        return (self.frequency == ScheduleFrequency.DAILY and
                self.interval == 1 and
                self.adjustment == BusinessDayAdjustment.NO_ADJUSTMENT)

    def __last_date(self):
        return self.start_date + relativedelta(years=+50)

    def is_due(self, test_date: date) -> bool:
        if self.__is_simple_daily_schedule():
            if self.end_type == ScheduleEndType.NO_END:
                return test_date >= self.start_date
            elif self.end_type == ScheduleEndType.END_DATE:
                return self.start_date <= test_date <= self.end_date

        if not self.cached_dates:
            # create dictionary of dates for each date
            self.cached_dates = {date: date for date in self.get_all_dates(self.__last_date())}

        # check if test_date is in dictionary
        return test_date in self.cached_dates

    def get_all_dates(self, to_date: date):
        if self.cached_dates:
            return self.cached_dates.keys()

        dates: list[date] = []
        repeats: int = 1
        iterator: date = self.start_date
        adjusted_date = self.__get_adjusted(iterator)

        while not self.__is_completed(repeats, adjusted_date, to_date):
            dates.append(adjusted_date)

            iterator = self.__next(repeats)
            adjusted_date = self.__get_adjusted(iterator)
            repeats += 1

        dates.extend(include_date for include_date in self.include_dates if include_date not in dates)

        dates = [keep_date for keep_date in dates if keep_date not in self.exclude_dates]

        self.cached_dates = {iter: iter for iter in dates}

        return dates

    def __next(self, repeats: int) -> date:
        if self.frequency == ScheduleFrequency.DAILY:
            return self.start_date + relativedelta(days=+ self.interval * repeats)

        return self.start_date + relativedelta(months=+ self.interval * repeats)

    def __is_completed(self, repeats: int, test_date: date, last_date: date) -> bool:
        if test_date > last_date:
            return True
        elif self.end_type == ScheduleEndType.END_DATE and test_date > self.end_date:
            return True
        elif self.end_type == ScheduleEndType.NO_END:
            return False

        return (self.end_type == ScheduleEndType.END_REPEATS and
                repeats > self.number_of_repeats)

    def __get_adjusted(self, test_date: date):
        if self.adjustment == BusinessDayAdjustment.NO_ADJUSTMENT:
            return test_date

        # TODO add calendar and call get_adjusted()
        return test_date


class ExternalTransaction(BaseModel):
    transaction_type_name: str
    amount: Decimal
    value_date: date


class PropertyValue(BaseModel):
    value: Dict[date, Any] = {}

    def __getitem__(self, value_date: date):
        # find first date that is less than or equal to value_date
        index_date = max([d for d in self.value.keys() if d <= value_date])
        return self.value[index_date]

    def __setitem__(self, value_date: date, value):
        self.value[value_date] = value


class Instalment(BaseModel):
    amount: Decimal
    is_fixed: bool


class Account(BaseModel):
    start_date: date
    account_type_name: str
    positions: dict[str, Position] = {}
    value_dated_properties: dict[str, PropertyValue] = {}
    properties: dict[str, Any] = {}
    dates: dict[str, date] = {}
    schedules: dict[str, Schedule] = {}
    transactions: list[Transaction] = []
    instalments: dict[str, Instalment] = {}

    def __init__(self, **kw):
        super().__init__(**kw)
        if "account_type" in kw:
            account_type: AccountType = kw["account_type"]
            self.__validate_properties(account_type)
            self.__initialize_positions(account_type)
            self.__initialize_schedules(account_type)
            self.__initialize_instalment(account_type)

    def __validate_properties(self, account_type: AccountType):
        for property_type in account_type.property_types:
            if property_type.value_dated:
                if property_type.name not in self.value_dated_properties:
                    raise ValueError(
                        f"Value dated property {property_type.name} is required for account type {account_type.name}")
            else:
                if property_type.name not in self.properties:
                    raise ValueError(f"Property {property_type.name} is required for account type {account_type.name}")

    def __initialize_schedules(self, account_type: AccountType):
        for schedule_type in account_type.schedule_types:
            # skip if schedule already exists
            if schedule_type.name in self.schedules:
                continue

            schedule = Schedule(
                start_date=self.evaluate(schedule_type.start_date_expression,
                                         {"accountType": account_type, "account": self,
                                          "value_date": self.start_date}),
                end_type=schedule_type.end_type,
                frequency=schedule_type.frequency,
                interval=self.evaluate(schedule_type.interval_expression, {"accountType": account_type,
                                                                           "account": self,
                                                                           "value_date": self.start_date}),
                adjustment=schedule_type.business_day_adjustment)

            if schedule_type.end_date_expression:
                schedule.end_date = self.evaluate(schedule_type.end_date_expression,
                                                  {"accountType": account_type,
                                                   "account": self,
                                                   "value_date": self.start_date})

            if schedule_type.number_of_repeats_expression:
                schedule.number_of_repeats = self.evaluate(schedule_type.number_of_repeats_expression)

            self.schedules[schedule_type.name] = schedule

    def __initialize_positions(self, account_type: AccountType):
        for transaction_type in account_type.transaction_types:
            for rule in transaction_type.position_rules:
                if rule.position_type_name not in self.positions:
                    self.positions[rule.position_type_name] = Position()

    def __initialize_instalment(self, account_type: AccountType):
        instalment_type = account_type.instalment_type

        # if no instalments create default one
        if instalment_type and len(self.instalments.items()) == 0:
            for date_value in self.schedules[instalment_type.schedule_name].get_all_dates(
                    self.start_date + relativedelta(years=+50)):
                self.instalments[date_value.strftime("%Y-%m-%d")] = Instalment(amount=Decimal(0), is_fixed=False)

    def apply_calculated_installment(self, amount: Decimal):
        # set all instalments to calculated amount if fixed is false
        for instalment in self.instalments.values():
            if not instalment.is_fixed:
                instalment.amount = amount

    def add_transaction(self, transaction: Transaction, transaction_type: TransactionType) -> dict[str, Decimal]:
        updated_positions: dict[str, Decimal] = {}
        for rule in transaction_type.position_rules:
            position = self.positions[rule.position_type_name]
            position.apply_operation(rule.operation, transaction.amount)
            updated_positions[rule.position_type_name] = position.amount

        self.transactions.append(transaction)

        return updated_positions

    def evaluate(self, expression: str, locals: Optional[Mapping[str, Any]]) -> Any:
        try:
            value = eval(expression, None, locals)
        except Exception as e:
            raise ValueError(f'Error evaluating expression: {expression} {e.args}') from e
        else:
            return value

    def __getattr__(self, method_name):
        if method_name in self.positions:
            return self.positions[method_name].amount
        if method_name in self.properties:
            return self.properties[method_name]
        if method_name in self.dates:
            return self.dates[method_name]
        else:
            raise AttributeError(f'No such attribute: {method_name}')

    class Config:
        exclude = {"config"}


def group_by_date(external_transactions: List[ExternalTransaction]) -> Dict[date, List[ExternalTransaction]]:
    keyfunc = lambda x: x.value_date

    groups = groupby(sorted(external_transactions, key=keyfunc), keyfunc)  # Sort the list by the keys and group by them

    return {key: list(group) for key, group in groups}


class TransactionDifference(BaseModel):
    amount = Decimal(0)
    value_date: date
    transaction_type: str
    amount: Decimal
    original: List[Transaction] = []
    new: List[Transaction] = []


def valuation_difference(original: List[Transaction], new: List[Transaction]) -> Dict[
    date, list[TransactionDifference]]:
    keyfunc = lambda x: (x.value_date, x.transaction_type)

    original_grouped = {key: list(group)
                        for key, group
                        in groupby(
            sorted(original, key=keyfunc), keyfunc)}

    new_grouped = {key: list(group)
                   for key, group
                   in groupby(
            sorted(new, key=keyfunc), keyfunc)}

    difference_list = list(get_difference(new_grouped, original_grouped))

    return {key: list(group)
            for key, group
            in groupby(
            sorted(difference_list, key=lambda x: x.value_date),
            lambda x: x.value_date)}


def get_difference(new_grouped, original_grouped):
    # Well done Chat GPT 4!
    for value_date, transaction_type in set(original_grouped.keys()).union(set(new_grouped.keys())):
        original_transactions = original_grouped.get((value_date, transaction_type), [])
        new_transactions = new_grouped.get((value_date, transaction_type), [])

        original_amount = sum([t.amount for t in original_transactions])
        new_amount = sum([t.amount for t in new_transactions])

        if original_amount != new_amount:
            yield TransactionDifference(
                value_date=value_date,
                transaction_type=transaction_type,
                amount=new_amount - original_amount,
                original=original_transactions,
                new=new_transactions)


class TransactionTrace(BaseModel):
    transaction: Transaction
    positions: Dict[str, Decimal]

    def __str__(self):
        return f" {self.transaction.value_date} {self.transaction.transaction_type} {self.transaction.amount} {self.positions}"


class AccountValuation(BaseModel):
    account: Account
    account_type: AccountType
    action_date: date
    trace: bool = False
    trace_list: List[TransactionTrace] = []

    def init_account(self):
        # reset all positions to zero
        for position in self.account.positions.values():
            position.amount = Decimal(0)

        self.account.transactions = []

        self.trace_list = []

    def forecast(self, to_value_date: date, external_transactions: dict[date, List[ExternalTransaction]]):
        value_date = self.account.start_date

        self.start_of_day(value_date)
        self.process_external_transactions(value_date, external_transactions)

        while value_date < to_value_date:
            self.end_of_day(value_date)

            value_date = value_date + timedelta(days=1)

            self.start_of_day(value_date)
            self.process_external_transactions(value_date, external_transactions)

    def process_external_transactions(self, value_date: date,
                                      external_transactions: dict[date, List[ExternalTransaction]]):
        if value_date in external_transactions:
            for external_transaction in external_transactions[value_date]:
                transaction_type = self.account_type.get_transaction_type(external_transaction.transaction_type_name)
                self.__create_transaction(transaction_type, value_date, external_transaction.amount, False)

    def start_of_day(self, value_date):
        for scheduled_transaction in self.account_type.scheduled_transactions:
            if scheduled_transaction.timing == ScheduledTransactionTiming.START_OF_DAY:
                self.__create_transaction_if_due(value_date, scheduled_transaction)

        value_date_str = value_date.strftime('%Y-%m-%d')

        if self.account_type.instalment_type:
            if self.account_type.instalment_type.timing == ScheduledTransactionTiming.START_OF_DAY \
                    and value_date_str in self.account.instalments:
                instalment = self.account.instalments[value_date_str]
                transaction_type = self.account_type.get_transaction_type(
                    self.account_type.instalment_type.transaction_type)
                self.__create_transaction(transaction_type, value_date, instalment.amount, True)

    def __create_transaction_if_due(self, value_date: date, scheduled_transaction: ScheduledTransaction):
        schedule = self.account.schedules[scheduled_transaction.schedule_name]

        if schedule.is_due(value_date):
            transaction_type = self.account_type.get_transaction_type(scheduled_transaction.generated_transaction_type)

            self.__create_calculated_transaction(value_date, transaction_type, scheduled_transaction.amount_expression)

    def __create_calculated_transaction(self, value_date: date, transaction_type: TransactionType,
                                        amount_expression: str):
        try:
            amount = self.account.evaluate(amount_expression,
                                           {"accountType": self.account_type,
                                            "account": self.account,
                                            "value_date": value_date})

            if not transaction_type.maximum_precision:
                amount = Decimal(round(amount, 2))

        except Exception as e:
            raise Exception(
                f'Error calculating {transaction_type.name} on {value_date} expression : {amount_expression} {e.args}') from e
        else:
            if amount != Decimal(0):
                self.__create_transaction(transaction_type, value_date, amount, True)

    def __create_transaction(self, transaction_type: TransactionType, value_date: date,
                             amount: Decimal, system_generated: bool):

        transaction = Transaction(action_date=self.action_date, value_date=value_date,
                                  transaction_type=transaction_type.name,
                                  amount=amount, system_generated=system_generated)

        positions = self.account.add_transaction(transaction, transaction_type)

        if self.trace:
            self.trace_list.append(TransactionTrace(transaction=transaction, positions=positions))

        triggered_transaction = self.account_type.get_trigger_transaction(transaction_type.name)

        if triggered_transaction:
            trigger_amount = self.account.evaluate(triggered_transaction.amount_expression,
                                                   {"transaction": transaction,
                                                    "accountType": self.account_type,
                                                    "account": self.account,
                                                    "value_date": value_date})

            generated_transaction_type = self.account_type.get_transaction_type(
                triggered_transaction.generated_transaction_type)
            self.__create_transaction(generated_transaction_type, value_date, trigger_amount, True)

    def end_of_day(self, value_date):
        for scheduled_transaction in self.account_type.scheduled_transactions:
            if scheduled_transaction.timing == ScheduledTransactionTiming.END_OF_DAY:
                self.__create_transaction_if_due(value_date, scheduled_transaction)

    def __calculate_for_instalment(self, value: Decimal) -> Decimal:
        self.init_account()

        self.account.apply_calculated_installment(value)
        self.forecast(self.account.dates[self.account_type.instalment_type.solve_for_date], {})
        result = self.account.positions[self.account_type.instalment_type.solve_for_zero_position].amount
        print(f'instalment {value} -> {self.account_type.instalment_type.solve_for_zero_position} {result}')
        return result

    def solve_instalment(self) -> Decimal:
        amount = scipy.optimize.brentq(self.__calculate_for_instalment, Decimal(-100000000), Decimal(100000000),
                                       xtol=Decimal(0.01))

        amount = Decimal(round(amount, 2))
        # apply amount to instalments
        self.account.apply_calculated_installment(amount)

        return amount


class Solver:
    valuer: AccountValuation

    def __init__(self, valuer: AccountValuation):
        self.valuer = valuer


class HolidayDate(BaseModel):
    description: str
    value: date


class BusinessDayCalculation(Enum):
    ANY_DAY = "AnyDay"
    PREVIOUS_BUSINESS_DAY = "PreviousBusinessDay"
    NEXT_BUSINESS_DAY = "NextBusinessDay"
    CLOSEST_BUSINESS_DAY_OR_NEXT = "ClosestBusinessDayOrNext"
    NEXT_BUSINESS_DAY_THIS_MONTH_OR_PREVIOUS = "NextBusinessDayThisMonthOrPrevious"


class Calendar(BaseModel):
    name: str
    is_default: bool
    holidays: List[HolidayDate] = []
    holidays_map: Dict[date, HolidayDate] = None

    def add(self, description: str, value: date) -> 'Calendar':
        self.holidays.append(HolidayDate(description=description, value=value))
        return self

    def __holidays_map(self):
        if self.holidays_map is None:
            self.holidays_map = {holiday.value: holiday for holiday in self.holidays}

        return self.holidays_map

    def is_business_day(self, value: date):
        if value.weekday() == 5 or value.weekday() == 6:
            return False

        return value not in self.__holidays_map()

    def get_calculated_business_day(self, value: date, adjustment: BusinessDayCalculation):
        if adjustment == BusinessDayCalculation.ANY_DAY:
            return value

        if adjustment == BusinessDayCalculation.PREVIOUS_BUSINESS_DAY:
            return self.get_previous_business_day(value)

        if adjustment == BusinessDayCalculation.NEXT_BUSINESS_DAY:
            return self.get_next_business_day(value)

        previous_business_day = self.get_previous_business_day(value)
        next_business_day = self.get_next_business_day(value)

        if adjustment == BusinessDayCalculation.CLOSEST_BUSINESS_DAY_OR_NEXT:
            if (value - previous_business_day).days > (next_business_day - value).days:
                return next_business_day
            elif (value - previous_business_day).days < (next_business_day - value).days:
                return previous_business_day
            else:
                return next_business_day

        # last option is NextBusinessDayThisMonthOrPrevious

        if next_business_day.month == value.month:
            return next_business_day

        return previous_business_day

    def get_previous_business_day(self, date: date):
        while not self.is_business_day(date):
            date = date - timedelta(days=1)

        return date

    def get_next_business_day(self, date: date):
        while not self.is_business_day(date):
            date = date + timedelta(days=1)

        return date



import unittest

from accounts.runtime import *
from tests.test_config import create_savings_account


def evaluate_account(account_type: AccountType, monthly_fee: Decimal, deposit: Decimal,
                     withholding_tax: Decimal) -> Account:
    start_date = date(2019, 1, 1)
    account = Account(start_date=start_date, account_type_name=account_type.name,
                      account_type=account_type,
                      properties={"monthlyFee": PropertyValue(value={start_date: monthly_fee}),
                                  "withholdingTax": PropertyValue(value={start_date: withholding_tax})})

    valuation = AccountValuation(account= account, account_type= account_type, action_date= date(2020, 1, 1))
    deposit_transaction_type = account_type.get_transaction_type("deposit")
    external_transactions = group_by_date([
        ExternalTransaction(transaction_type_name=deposit_transaction_type.name,
                            amount=deposit, value_date=date(2019, 1, 1))])

    valuation.forecast(date(2020, 1, 1), external_transactions)
    return account


def evaluate_account_with_withholding_tax_change(account_type: AccountType, monthly_fee: Decimal, deposit: Decimal,
                                                 withholding_tax1: Decimal, withholding_tax2: Decimal,
                                                 tax_change_date: date) -> Account:
    start_date = date(2019, 1, 1)
    account = Account(start_date=start_date, account_type_name=account_type.name,
                      account_type=account_type,
                      properties={"monthlyFee": PropertyValue(value={start_date: monthly_fee}),
                                  "withholdingTax": PropertyValue(value={start_date: withholding_tax1,
                                                                         tax_change_date: withholding_tax2})})

    valuation = AccountValuation(account=account, account_type=account_type, action_date=date(2020, 1, 1))
    deposit_transaction_type = account_type.get_transaction_type("deposit")
    external_transactions = group_by_date([
        ExternalTransaction(transaction_type_name=deposit_transaction_type.name,
                            amount=deposit, value_date=date(2019, 1, 1))])

    valuation.forecast(date(2020, 1, 1), external_transactions)
    return account


class TestAccount(unittest.TestCase):
    def test_create_account(self):
        account_type = create_savings_account()

        account = Account(start_date=date.today(),
                          account_type_name=account_type.name,
                          account_type=account_type,
                          properties={"monthlyFee": Decimal(1),
                                      "withholdingTax": Decimal(0.2)})

        self.assertEqual(3, len(account.positions.keys()))

        json = account.json()

        account2 = Account.parse_raw(json)

        self.assertEqual(account.start_date, account2.start_date)

    def test_eval(self):
        account_type = create_savings_account()

        account = Account(start_date=date(2019, 1, 1),
                          account_type_name=account_type.name,
                          account_type=account_type,
                          properties={"monthlyFee": Decimal(1),
                                      "withholdingTax": Decimal(0.2)}
                          )

        self.assertEqual(Decimal(0), account.evaluate("self.current", None))

        self.assertEqual(date(2019, 1, 1), account.evaluate("self.start_date", None))

        self.assertEqual(date(2019, 1, 1) + relativedelta(months=+1) + relativedelta(days=-1),
                         account.evaluate("self.start_date + relativedelta(months=+1) + relativedelta(days=-1)",
                                          None))

    def test_valuation(self):
        account_type = create_savings_account()

        account = evaluate_account(account_type, monthly_fee=Decimal(1), deposit=Decimal(1000),
                                   withholding_tax=Decimal(0.2))

        self.assertAlmostEqual(account.positions['current'].amount, Decimal(1018.25), places=4)
        self.assertAlmostEqual(account.positions['withholding'].amount, Decimal(6.05), places=4)
        self.assertAlmostEqual(account.transactions[1].amount, Decimal(0.08219), places=4)

    def test_valuation_difference(self):
        account_type = create_savings_account()

        original = evaluate_account(account_type, monthly_fee=Decimal(1),
                                    deposit=Decimal(1000), withholding_tax=Decimal(0.2)).transactions
        new = evaluate_account(account_type, monthly_fee=Decimal(0),
                               deposit=Decimal(1000), withholding_tax=Decimal(0.1)).transactions

        difference = valuation_difference(original, new)

        self.assertEqual(335, len(difference.keys()))

        diff: List[TransactionDifference] = difference[date(2019, 1, 31)]

        self.assertEqual(3, len(diff))

        fee = next(t for t in diff if t.transaction_type == 'fee')
        withholding = next(t for t in diff if t.transaction_type == 'withholdingTax')

        self.assertEqual(Decimal(-1), fee.amount)
        self.assertEqual(1, len(fee.original))
        self.assertEqual(0, len(fee.new))
        self.assertAlmostEqual(Decimal(-0.26), withholding.amount, 2)

    def test_property_valuation(self):
        account_type = create_savings_account()

        account = evaluate_account_with_withholding_tax_change(account_type, monthly_fee=Decimal(1),
                                                               deposit=Decimal(1000),
                                                               withholding_tax1=Decimal(0.2),
                                                               withholding_tax2=Decimal(0.1),
                                                               tax_change_date=date(2019, 7, 1))

        self.assertAlmostEqual(account.positions['current'].amount, Decimal(1018.25), places=2)
        self.assertAlmostEqual(account.positions['withholding'].amount, Decimal(4.52), places=2)
        self.assertAlmostEqual(account.transactions[1].amount, Decimal(0.08219), places=4)


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import date
from decimal import Decimal

from dateutil.relativedelta import relativedelta
from accounts.metadata import AccountType
from accounts.runtime import Account, PropertyValue, AccountValuation, Schedule
from tests.test_config import create_loan_given_account


def create_without_mandatory_properties():
    account_type = create_loan_given_account()

    start_date = date(2013, 3, 8)
    end_date = start_date + relativedelta(years=+25)

    account = Account(start_date=start_date,
                      account_type_name=account_type.name,
                      account_type=account_type,
                      dates={"accrual_start": start_date,
                             "end_date": end_date}
                      )


def create_loan_account(account_type: AccountType, start_date: date) -> (Account, date):
    end_date = start_date + relativedelta(years=+25)
    dates = {"accrual_start": start_date,
             "end_date": end_date}
    properties = {"advance": 624000,
                  "payment": 0}

    # get initial values
    account = Account(start_date=start_date,
                      account_type_name=account_type.name,
                      account_type=account_type,
                      dates=dates,
                      properties=properties
                      )

    interest_start = date(2013, 3, 31)
    interest_schedule: Schedule = account.schedules["interest"]
    interest_schedule.cached_dates = {}
    interest_schedule.start_date = interest_start
    interest_schedule.end_date = end_date
    interest_schedule.include_dates.append(end_date)
    redemption_schedule: Schedule = account.schedules["redemption"]
    redemption_schedule.cached_dates = {}
    redemption_schedule.start_date = interest_start
    redemption_schedule.end_date = end_date
    redemption_schedule.include_dates.append(end_date)

    # create new account using schedules above

    account = Account(start_date=start_date,
                      account_type_name=account_type.name,
                      account_type=account_type,
                      dates=dates,
                      properties=properties,
                      schedules={"interest": interest_schedule,
                                 "redemption": redemption_schedule})

    dict = account.dict()
    json = account.json()

    return account, end_date


class TestConfiguration(unittest.TestCase):

    def test_serialize(self):
        account_type = None

        with open('loan_given.json', 'r') as f:
            json_str = f.read()
            account_type = AccountType.parse_raw(json_str)

        start_date = date(2013, 3, 8)
        end_date = start_date + relativedelta(year=+25)

        parameters = {"advance_amount": 624000}

        account = Account(start_date=start_date,
                          account_type_name=account_type.name,
                          account_type=account_type,
                          dates={"accrual_start": start_date,
                                 "end_date": end_date}
                          )

        self.assertEqual(3, len(account.positions.keys()))

    def test_serialize_to_json(self):
        account_type = create_loan_given_account()

        json = account_type.json()

        self.assertIsNotNone(json)

        account_type2 = AccountType.parse_raw(json)

        json2 = account_type2.json()

        self.assertEqual(json, json2)

    def test_mandatory_properties(self):
        self.assertRaises(ValueError, create_without_mandatory_properties)

    def test_forecast(self):
        account_type = create_loan_given_account()

        account, end_date = create_loan_account(account_type, date(2013, 3, 8))

        valuation = AccountValuation(account=account, account_type=account_type, action_date=end_date, trace=True)

        valuation.forecast(end_date + relativedelta(days=1), [])

        self.assertEqual(5, len(account.positions.keys()))
        self.assertAlmostEqual(Decimal(1333778.93), account.positions["principal"].amount, places=2)
        self.assertAlmostEqual(Decimal(709778.93), account.positions["interest_capitalized"].amount, places=2)
        self.assertAlmostEqual(Decimal(0.005), account.positions["accrued"].amount, places=2)

    def test_installments(self):
        account_type = create_loan_given_account()

        account, end_date = create_loan_account(account_type, date(2013, 3, 8))

        valuation = AccountValuation(account=account, account_type=account_type, action_date=end_date, trace=False)

        payment = valuation.solve_instalment()

        self.assertAlmostEqual(Decimal(2964.37), Decimal(payment), places=2)



from accounts.metadata import *
from accounts.runtime import Calendar


def create_savings_account() -> AccountType:
    acc = AccountType(name="savingsAccount", label="Savings Account")

    current = acc.add_position_type("current", "current balance")
    interest_accrued = acc.add_position_type("accrued", "interest accrued")
    withholding = acc.add_position_type("withholding", "withholding tax")

    acc.add_property_type("monthlyFee", "Monthly Fee", DataType.DECIMAL, True)
    acc.add_property_type("withholdingTax", "Withholding Tax Rate", DataType.DECIMAL, True)

    acc.add_transaction_type("deposit", "Deposit") \
        .add_position_rule(TransactionOperation.CREDIT, current)

    fee_tt = acc.add_transaction_type("fee", "Fee") \
        .add_position_rule(TransactionOperation.DEBIT, current)

    interest_accrued_tt = acc.add_transaction_type("interestAccrued", "Interest Accrued", True) \
        .add_position_rule(TransactionOperation.CREDIT, interest_accrued)

    capitalized_tt = acc.add_transaction_type("capitalized", "Interest Capitalized") \
        .add_position_rule(TransactionOperation.CREDIT, current) \
        .add_position_rule(TransactionOperation.DEBIT, interest_accrued)

    withholding_tt = acc.add_transaction_type("withholdingTax", "Withholding Tax") \
        .add_position_rule(TransactionOperation.CREDIT, withholding)

    accrual_schedule = ScheduleType(name="accrual", label="Accrual Schedule", frequency=ScheduleFrequency.DAILY,
                                    end_type=ScheduleEndType.NO_END,
                                    business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                    interval_expression="1", start_date_expression="account.start_date")

    acc.add_schedule_type(accrual_schedule)

    compounding_schedule = ScheduleType(name="compounding", label="Compounding Schedule",
                                        frequency=ScheduleFrequency.MONTHLY,
                                        end_type=ScheduleEndType.NO_END,
                                        business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                        interval_expression="1",
                                        start_date_expression="account.start_date + relativedelta(month=+1) + relativedelta(days=-1)")

    acc.add_schedule_type(compounding_schedule)

    acc.add_scheduled_transaction(compounding_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                  fee_tt,
                                  "account.monthlyFee[value_date]")

    acc.add_scheduled_transaction(accrual_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                  interest_accrued_tt,
                                  "account.current * accountType.interest.get_rate(value_date, account.current) / Decimal(365)")

    acc.add_scheduled_transaction(compounding_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                  capitalized_tt, "account.accrued")

    interest_rate = acc.add_rate_type("interest", "Interest Rate")

    interest_rate.add_tier(date(2019, 1, 1), Decimal(10000), Decimal(0.03))
    interest_rate.add_tier(date(2019, 1, 1), Decimal(100000), Decimal(0.035))
    interest_rate.add_tier(date(2019, 1, 1), Decimal(50000), Decimal(0.04))

    acc.add_trigger_transaction(capitalized_tt, withholding_tt,
                                "transaction.amount * account.withholdingTax[value_date]")

    return acc


def create_loan_given_account() -> AccountType:
    loan_given = AccountType(name="Loan", label="Loan")

    conversion_interest_position = loan_given.add_position_type("conversion_interest", "Conversion Interest")
    early_redemption_fee_position = loan_given.add_position_type("early_redemption_fee", "Early Redemption Fee")
    interest_accrued_position = loan_given.add_position_type("accrued", "Interest Accrued")
    interest_capitalized_position = loan_given.add_position_type("interest_capitalized", "Interest Capitalized")
    principal_position = loan_given.add_position_type("principal", "Principal")

    loan_given.add_date_type(name="accrual_start", label="Accrual Start Date")
    loan_given.add_date_type(name="end_date", label="End Date")

    accrual_schedule = ScheduleType(name="accrual", label="Accrual Schedule", frequency=ScheduleFrequency.DAILY,
                                    end_type=ScheduleEndType.NO_END,
                                    business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                    interval_expression="1", start_date_expression="account.start_date")

    interest_schedule = ScheduleType(name="interest", label="Interest Schedule", frequency=ScheduleFrequency.MONTHLY,
                                     end_type=ScheduleEndType.NO_END,
                                     business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                     interval_expression="1", start_date_expression="account.start_date",
                                     end_date_expression="account.end_date",
                                     include_dates_expression="account.end_date")

    redemption_schedule = ScheduleType(name="redemption", label="Redemption Schedule",
                                       frequency=ScheduleFrequency.MONTHLY,
                                       end_type=ScheduleEndType.NO_END,
                                       business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                       interval_expression="1",
                                       start_date_expression="account.start_date + relativedelta(months=+1)",
                                       end_date_expression="account.end_date",
                                       include_dates_expression="account.end_date")

    advance_schedule = ScheduleType(name="advance", label="Advance Schedule",
                                    frequency=ScheduleFrequency.DAILY,
                                    end_type=ScheduleEndType.END_DATE,
                                    business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                    interval_expression="1",
                                    start_date_expression="account.start_date",
                                    end_date_expression="account.start_date")

    loan_given.add_schedule_type(accrual_schedule)
    loan_given.add_schedule_type(interest_schedule)
    loan_given.add_schedule_type(redemption_schedule)
    loan_given.add_schedule_type(advance_schedule)

    interest_accrued = loan_given.add_transaction_type("interestAccrued", "Interest Accrued", True) \
        .add_position_rule(TransactionOperation.CREDIT, interest_accrued_position)

    interest_capitalized = loan_given.add_transaction_type("interestCapitalized", "Interest Capitalized") \
        .add_position_rule(TransactionOperation.CREDIT, interest_capitalized_position) \
        .add_position_rule(TransactionOperation.DEBIT, interest_accrued_position) \
        .add_position_rule(TransactionOperation.CREDIT, principal_position)

    early_redemption_fee = loan_given.add_transaction_type("earlyRedemptionFee", "Early Redemption Fee") \
        .add_position_rule(TransactionOperation.CREDIT, early_redemption_fee_position)

    conversion_interest = loan_given.add_transaction_type("conversionInterest", "Conversion Interest") \
        .add_position_rule(TransactionOperation.CREDIT, conversion_interest_position)

    redemption = loan_given.add_transaction_type("redemption", "Redemption") \
        .add_position_rule(TransactionOperation.DEBIT, principal_position)

    advance_transaction = loan_given.add_transaction_type("advance", "Advance") \
        .add_position_rule(TransactionOperation.CREDIT, principal_position)

    additional_advance_transaction = loan_given.add_transaction_type("additionalAdvance", "Additional Advance") \
        .add_position_rule(TransactionOperation.CREDIT, principal_position)

    interest_payment_transaction = loan_given.add_transaction_type("interestPayment", "Interest Payment") \
        .add_position_rule(TransactionOperation.DEBIT, interest_accrued_position)

    loan_given.add_scheduled_transaction(accrual_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                         interest_accrued,
                                         "account.principal * accountType.interest.get_rate(value_date, "
                                         "account.principal) / Decimal(365)")

    loan_given.add_scheduled_transaction(interest_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                         interest_capitalized,
                                         "account.accrued")

    loan_given.add_scheduled_transaction(advance_schedule, ScheduledTransactionTiming.START_OF_DAY,
                                         advance_transaction,
                                         "account.advance")

    loan_given.add_instalment_type(name="payments", label="Payments", timing=ScheduledTransactionTiming.START_OF_DAY,
                                   transaction_type=redemption.name,
                                   property_name="payment",
                                   solve_for_zero_position="principal",
                                   solve_for_date="end_date",
                                   schedule_name=redemption_schedule.name)

    loan_given.add_property_type("advance", "Advance Amount", DataType.DECIMAL, True)
    loan_given.add_property_type("payment", "Payment Amount", DataType.DECIMAL, True)

    interest_rate = loan_given.add_rate_type("interest", "Interest Rate")

    interest_rate.add_tier(date(2000, 1, 1), Decimal(2000000), Decimal(0.0304))
    interest_rate.add_tier(date(2000, 1, 1), Decimal(10000000), Decimal(0.025))
    interest_rate.add_tier(date(2000, 1, 1), Decimal(1E30), Decimal(0.02))

    return loan_given


def get_euro_calendar():
    return Calendar(name="Euro Calendar", is_default=True) \
        .add("GOOD FRIDAY", date(2000, 4, 21)) \
        .add("EASTER MONDAY", date(2000, 4, 24)) \
        .add("LABOUR DAY (01 MAY)", date(2000, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2000, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2000, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2001, 1, 1)) \
        .add("GOOD FRIDAY", date(2001, 4, 13)) \
        .add("EASTER MONDAY", date(2001, 4, 16)) \
        .add("LABOUR DAY (01 MAY)", date(2001, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2001, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2001, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2002, 1, 1)) \
        .add("GOOD FRIDAY", date(2002, 3, 29)) \
        .add("EASTER MONDAY", date(2002, 4, 1)) \
        .add("LABOUR DAY (01 MAY)", date(2002, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2002, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2002, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2003, 1, 1)) \
        .add("GOOD FRIDAY", date(2003, 4, 18)) \
        .add("EASTER MONDAY", date(2003, 4, 21)) \
        .add("LABOUR DAY (01 MAY)", date(2003, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2003, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2003, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2004, 1, 1)) \
        .add("GOOD FRIDAY", date(2004, 4, 9)) \
        .add("EASTER MONDAY", date(2004, 4, 12)) \
        .add("GOOD FRIDAY", date(2005, 3, 25)) \
        .add("EASTER MONDAY", date(2005, 3, 28)) \
        .add("BOXING DAY (26 DEC)", date(2005, 12, 26)) \
        .add("GOOD FRIDAY", date(2006, 4, 14)) \
        .add("EASTER MONDAY", date(2006, 4, 17)) \
        .add("LABOUR DAY (01 MAY)", date(2006, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2006, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2006, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2007, 1, 1)) \
        .add("GOOD FRIDAY", date(2007, 4, 6)) \
        .add("EASTER MONDAY", date(2007, 4, 9)) \
        .add("LABOUR DAY (01 MAY)", date(2007, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2007, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2007, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2008, 1, 1)) \
        .add("GOOD FRIDAY", date(2008, 3, 21)) \
        .add("EASTER MONDAY", date(2008, 3, 24)) \
        .add("LABOUR DAY (01 MAY)", date(2008, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2008, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2008, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2009, 1, 1)) \
        .add("GOOD FRIDAY", date(2009, 4, 10)) \
        .add("EASTER MONDAY", date(2009, 4, 13)) \
        .add("LABOUR DAY (01 MAY)", date(2009, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2009, 12, 25)) \
        .add("NEW YEARS DAY (01JAN)", date(2010, 1, 1)) \
        .add("GOOD FRIDAY", date(2010, 4, 2)) \
        .add("EASTER MONDAY", date(2010, 4, 5)) \
        .add("GOOD FRIDAY", date(2011, 4, 22)) \
        .add("EASTER MONDAY", date(2011, 4, 25)) \
        .add("BOXING DAY (26 DEC)", date(2011, 12, 26)) \
        .add("GOOD FRIDAY", date(2012, 4, 6)) \
        .add("EASTER MONDAY", date(2012, 4, 9)) \
        .add("LABOUR DAY (01 MAY)", date(2012, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2012, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2012, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2013, 1, 1)) \
        .add("GOOD FRIDAY", date(2013, 3, 29)) \
        .add("EASTER MONDAY", date(2013, 4, 1)) \
        .add("LABOUR DAY (01 MAY)", date(2013, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2013, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2013, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2014, 1, 1)) \
        .add("GOOD FRIDAY", date(2014, 4, 18)) \
        .add("EASTER MONDAY", date(2014, 4, 21)) \
        .add("LABOUR DAY (01 MAY)", date(2014, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2014, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2014, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2015, 1, 1)) \
        .add("GOOD FRIDAY", date(2015, 4, 3)) \
        .add("EASTER MONDAY", date(2015, 4, 6)) \
        .add("LABOUR DAY (01 MAY)", date(2015, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2015, 12, 25)) \
        .add("NEW YEARS DAY (01JAN)", date(2016, 1, 1)) \
        .add("GOOD FRIDAY", date(2016, 3, 25)) \
        .add("EASTER MONDAY", date(2016, 3, 28)) \
        .add("BOXING DAY (26 DEC)", date(2016, 12, 26)) \
        .add("GOOD FRIDAY", date(2017, 4, 14)) \
        .add("EASTER MONDAY", date(2017, 4, 17)) \
        .add("LABOUR DAY (01 MAY)", date(2017, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2017, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2017, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2018, 1, 1)) \
        .add("GOOD FRIDAY", date(2018, 3, 30)) \
        .add("EASTER MONDAY", date(2018, 4, 2)) \
        .add("LABOUR DAY (01 MAY)", date(2018, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2018, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2018, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2019, 1, 1)) \
        .add("GOOD FRIDAY", date(2019, 4, 19)) \
        .add("EASTER MONDAY", date(2019, 4, 22)) \
        .add("LABOUR DAY (01 MAY)", date(2019, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2019, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2019, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2020, 1, 1)) \
        .add("GOOD FRIDAY", date(2020, 4, 10)) \
        .add("EASTER MONDAY", date(2020, 4, 13)) \
        .add("LABOUR DAY (01 MAY)", date(2020, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2020, 12, 25)) \
        .add("NEW YEARS DAY (01JAN)", date(2021, 1, 1)) \
        .add("GOOD FRIDAY", date(2021, 4, 2)) \
        .add("EASTER MONDAY", date(2021, 4, 5)) \
        .add("GOOD FRIDAY", date(2022, 4, 15)) \
        .add("EASTER MONDAY", date(2022, 4, 18)) \
        .add("BOXING DAY (26 DEC)", date(2022, 12, 26)) \
        .add("GOOD FRIDAY", date(2023, 4, 7)) \
        .add("EASTER MONDAY", date(2023, 4, 10)) \
        .add("LABOUR DAY (01 MAY)", date(2023, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2023, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2023, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2024, 1, 1)) \
        .add("GOOD FRIDAY", date(2024, 3, 29)) \
        .add("EASTER MONDAY", date(2024, 4, 1)) \
        .add("LABOUR DAY (01 MAY)", date(2024, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2024, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2024, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2025, 1, 1)) \
        .add("GOOD FRIDAY", date(2025, 4, 18)) \
        .add("EASTER MONDAY", date(2025, 4, 21)) \
        .add("LABOUR DAY (01 MAY)", date(2025, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2025, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2025, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2026, 1, 1)) \
        .add("GOOD FRIDAY", date(2026, 4, 3)) \
        .add("EASTER MONDAY", date(2026, 4, 6)) \
        .add("LABOUR DAY (01 MAY)", date(2026, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2026, 12, 25)) \
        .add("NEW YEARS DAY (01JAN)", date(2027, 1, 1)) \
        .add("GOOD FRIDAY", date(2027, 3, 26)) \
        .add("EASTER MONDAY", date(2027, 3, 29)) \
        .add("GOOD FRIDAY", date(2028, 4, 14)) \
        .add("EASTER MONDAY", date(2028, 4, 17)) \
        .add("LABOUR DAY (01 MAY)", date(2028, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2028, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2028, 12, 26)) \
        .add("NEW YEARS DAY (01JAN)", date(2029, 1, 1)) \
        .add("GOOD FRIDAY", date(2029, 3, 30)) \
        .add("EASTER MONDAY", date(2029, 4, 2)) \
        .add("LABOUR DAY (01 MAY)", date(2029, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2029, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2029, 12, 26)) \
        .add("NEW YEARS DAY (1 JAN)", date(2030, 1, 1)) \
        .add("GOOD FRIDAY", date(2030, 4, 19)) \
        .add("EASTER MONDAY", date(2030, 4, 22)) \
        .add("LABOUR DAY", date(2030, 5, 1)) \
        .add("CHRISTMAS DAY (25 DEC)", date(2030, 12, 25)) \
        .add("BOXING DAY (26 DEC)", date(2030, 12, 26))



import unittest
import yaml
from accounts.metadata import *
from tests.test_config import create_savings_account


class TestConfiguration(unittest.TestCase):
    def test_serialize(self):
        account_type = create_savings_account()

        # serialize config to text using Pydantic

        text = account_type.json()

        account_type2 = AccountType.parse_raw(text)

        text2 = account_type2.json()

        self.assertEqual(text, text2)


class RateTest(unittest.TestCase):
    def test_get_max_when_empy(self):
        rate_type = RateType(name='rates', label='Rates')

        max_value = rate_type.get_max_to_amount(value_date=date(2019, 7, 1))

        self.assertEqual(Decimal(0), max_value)

    def test_get_max_when_non_empy(self):
        rate_type = RateType(name='rates', label='Rates')

        rate_type.add_tier(date(2019, 1, 1), Decimal(100), Decimal(5))

        max_value = rate_type.get_max_to_amount(date(2019, 1, 1))

        self.assertEqual(Decimal(100), max_value)


if __name__ == '__main__':
    unittest.main()



from typing import Any


class CustomEncoder:
    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            return {str(k): self(v) for k, v in obj.items()}
        return obj



import unittest
from datetime import date

from accounts.runtime import BusinessDayCalculation
from tests.test_config import get_euro_calendar


class TestCalendar(unittest.TestCase):

    def test_long_weekend(self):
        calendar = get_euro_calendar()

        self.assertTrue(calendar.is_business_day(date(2019, 4, 18)))
        self.assertFalse(calendar.is_business_day(date(2019, 4, 19)))  # easter friday
        self.assertFalse(calendar.is_business_day(date(2019, 4, 20)))  # saturday
        self.assertFalse(calendar.is_business_day(date(2019, 4, 21)))  # sunday
        self.assertFalse(calendar.is_business_day(date(2019, 4, 22)))  # eastern monday
        self.assertTrue(calendar.is_business_day(date(2019, 4, 23)))

        self.assertEqual(date(2019, 4, 23), calendar.get_next_business_day(date(2019, 4, 19)))
        self.assertEqual(date(2019, 4, 18), calendar.get_previous_business_day(date(2019, 4, 22)))

        self.assertEqual(date(2019, 4, 23),
                         calendar.get_calculated_business_day(date(2019, 4, 19),
                                                              BusinessDayCalculation.NEXT_BUSINESS_DAY))
        self.assertEqual(date(2019, 4, 18),
                         calendar.get_calculated_business_day(date(2019, 4, 22),
                                                              BusinessDayCalculation.PREVIOUS_BUSINESS_DAY))

        self.assertEqual(date(2019, 4, 18),
                         calendar.get_calculated_business_day(date(2019, 4, 20),
                                                              BusinessDayCalculation.CLOSEST_BUSINESS_DAY_OR_NEXT))
        self.assertEqual(date(2019, 4, 23),
                         calendar.get_calculated_business_day(date(2019, 4, 21),
                                                              BusinessDayCalculation.CLOSEST_BUSINESS_DAY_OR_NEXT))

        self.assertEqual(date(2019, 8, 30),
                         calendar.get_calculated_business_day(date(2019, 8, 31),
                                                              BusinessDayCalculation.
                                                              NEXT_BUSINESS_DAY_THIS_MONTH_OR_PREVIOUS))

        self.assertEqual(date(2019, 9, 30),
                         calendar.get_calculated_business_day(date(2019, 9, 29),
                                                              BusinessDayCalculation.
                                                              NEXT_BUSINESS_DAY_THIS_MONTH_OR_PREVIOUS))

        self.assertEqual(date(2019, 9, 29),
                         calendar.get_calculated_business_day(date(2019, 9, 29),
                                                              BusinessDayCalculation.
                                                              ANY_DAY))  # no adjustment, non-working day is ok



from setuptools import setup, find_packages

setup(
    name='transaction-accounts',
    version='0.2.0',
    description='Create configuration for transactional accounts and implement account runtime',
    long_description='''
Transaction Accounts
====================

This library provides basic functionality for working with transaction accounts.

You can use it to make any type of transaction account, such as a savings account, a credit card, or a loan.

Assume we would like to create simple Savings Account that has 3 types of balances:

- current balance
- interest accrued
- withholding tax

We would like to have 4 types of transactions:

- deposit
- interest accrued
- interest capitalized
- withholding tax

We would like to have 2 types of schedules:

- accrual schedule
- compounding schedule

We would like to have interest rate with 3 tiers:

- 0 - 10000: 3%
- 10000 - 50000: 3.5%
- 50000+: 4%

We would like to have monthly fee of 1.00 charged at the end of each month to the account before interest is accrued.

Deposit transaction will increase current balance, and it will be used to deposit money to the account. 
This transaction will be externally created and posted to the account.

Interest accrued transaction will increase interest accrued balance, and it will be used to accrue interest on the account.

Interest will be accrued daily, and it will be calculated based on current balance and interest rate.

Interest capitalized transaction will increase current balance and decrease interest accrued balance.
This transaction will be internally created and posted to the account at the end of each month.

Withholding tax transaction will decrease withholding tax balance, and it will be used to pay withholding tax on the account. 
It will be calculated as 20% of interest capitalized transaction.

We will use accrual schedule to accrue interest daily, and we will use compounding schedule to capitalize interest at the end of each month.

We will use interest rate to calculate interest.

We will use trigger transaction to calculate withholding tax.

We will use monthly fee property to specify monthly fee. This amount will be charged at the end of each month.

Note that this framework is not limited to this configuration, and it can be used to create any type of transaction account.

You can define custom calculations when calculating either schedules or transactions. In this context you can use any of the following variables:

- account: Account
- transaction: Transaction
- config: Configuration

Here is a code that will create this configuration:

.. code:: python

 def create_savings_account() -> AccountType:
    acc = AccountType(name="savingsAccount", label="Savings Account")

    current = acc.add_position_type("current", "current balance")
    interest_accrued = acc.add_position_type("accrued", "interest accrued")
    withholding = acc.add_position_type("withholding", "withholding tax")

    montly_fee = acc.add_property_type("monthlyFee", "Monthly Fee", DataType.DECIMAL, True)

    acc.add_transaction_type("deposit", "Deposit").add_position_rule(TransactionOperation.CREDIT, current)

    fee_tt = acc.add_transaction_type("fee", "Fee").add_position_rule(TransactionOperation.DEBIT, current)

    interest_accrued_tt = acc.add_transaction_type("interestAccrued", "Interest Accrued").add_position_rule(TransactionOperation.CREDIT, interest_accrued)

    capitalized_tt = acc.add_transaction_type("capitalized", "Interest Capitalized").add_position_rule(TransactionOperation.CREDIT, current).add_position_rule(TransactionOperation.DEBIT, interest_accrued)

    withholding_tt = acc.add_transaction_type("withholdingTax", "Withholding Tax").add_position_rule(TransactionOperation.CREDIT, withholding)

    accrual_schedule = ScheduleType(name="accrual", label="Accrual Schedule", frequency=ScheduleFrequency.DAILY,
                                    end_type=ScheduleEndType.NO_END,
                                    business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                    interval_expression="1", start_date_expression="account.start_date")

    acc.add_schedule_type(accrual_schedule)

    compounding_schedule = ScheduleType(name="compounding", label="Compounding Schedule",
                                        frequency=ScheduleFrequency.MONTHLY,
                                        end_type=ScheduleEndType.NO_END,
                                        business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                        interval_expression="1",
                                        start_date_expression="account.start_date + relativedelta(month=+1) + relativedelta(days=-1)")

    acc.add_schedule_type(compounding_schedule)

    acc.add_scheduled_transaction(compounding_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                  fee_tt,
                                  "account.monthlyFee")

    acc.add_scheduled_transaction(accrual_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                  interest_accrued_tt,
                                  "account.current * accountType.interest.get_rate(account.current) / Decimal(365)")

    acc.add_scheduled_transaction(compounding_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                  capitalized_tt, "account.accrued")

    interest_rate = acc.add_rate_type("interest", "Interest Rate")

    interest_rate.add_tier(Decimal(10000), Decimal(0.03))
    interest_rate.add_tier(Decimal(100000), Decimal(0.035))
    interest_rate.add_tier(Decimal(50000), Decimal(0.04))

    acc.add_trigger_transaction(capitalized_tt, withholding_tt, "transaction.amount * Decimal(0.2)")

    return acc



Given configuration, we can create an account:

.. code:: python

 def test_valuation(self):
        account_type = create_savings_account()

        account = Account(start_date=date(2019, 1, 1), account_type_name=account_type.name,
                          account_type=account_type, properties={"monthlyFee": Decimal(1.00)})

        valuation = AccountValuation(account, account_type, date(2020, 1, 1))

        deposit_transaction_type = account_type.get_transaction_type("deposit")

        external_transactions = group_by_date([
            ExternalTransaction(transaction_type_name=deposit_transaction_type.name,
                                amount=Decimal(1000), value_date=date(2019, 1, 1))])

        valuation.forecast(date(2020, 1, 1), external_transactions)

        self.assertAlmostEqual(account.positions['current'].amount, Decimal(1018.24775), places=4)
        self.assertAlmostEqual(account.positions['withholding'].amount, Decimal(6.04955), places=4)
        self.assertAlmostEqual(account.transactions[1].amount, Decimal('0.08219'), places=4)
    
    ''',
    author='Igor Music',
    author_email='igormusich@gmail.com',
    packages=find_packages(),
    license='MIT',
    url='https://github.com/igormusic/transaction-accounts',
    download_url='https://github.com/igormusic/transaction-accounts/archive/refs/tags/0.0.6.tar.gz',
    keywords=['TRANSACTION PROCESSING', 'LOANS', 'SAVINGS', 'ACCOUNTS', 'FINANCE', 'BANKING'],
    install_requires=[
        'python-dateutil',
        'pydantic',
        'scipy'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
    ],
)



import unittest

from accounts.runtime import *


class TestSchedule(unittest.TestCase):
    start_date = date(2013, 3, 8)
    end_date = start_date + relativedelta(years=+25)

    def __get_daily_schedule(self, start_date: date, end_date: date, end_type: ScheduleEndType):
        return Schedule(start_date=start_date, end_type=end_type, frequency=ScheduleFrequency.DAILY, interval=1,
                        adjustement=BusinessDayAdjustment.NO_ADJUSTMENT, end_date=end_date)

    def test_daily_schedule_no_end_tests(self):
        accrual_schedule = self.__get_daily_schedule(self.start_date, self.end_date, ScheduleEndType.NO_END)

        accrual_dates = accrual_schedule.get_all_dates(self.end_date);

        self.assertEqual(self.start_date, accrual_dates[0])
        self.assertEqual(self.end_date, accrual_dates[len(accrual_dates) - 1])
        self.assertEqual((self.end_date - self.start_date).days + 1, len(accrual_dates))

        self.assertEqual(accrual_schedule.is_due(self.start_date), True)
        self.assertEqual(accrual_schedule.is_due(self.end_date), True)

        self.assertEqual(accrual_schedule.is_due(self.start_date + relativedelta(days=-1)), False)
        self.assertEqual(accrual_schedule.is_due(self.end_date + relativedelta(days=+1)), True)

    def test_monthly_schedule(self):
        interest_start = date(2013, 3, 31)
        schedule = Schedule(start_date=interest_start, end_type=ScheduleEndType.END_DATE,
                            frequency=ScheduleFrequency.MONTHLY,
                            interval=1, end_date=self.end_date)

        schedule.exclude_dates.append(date(2013, 12, 31))
        schedule.include_dates.append(self.end_date)

        interest_dates = schedule.get_all_dates(self.end_date)

        self.assertEqual(schedule.adjustment, BusinessDayAdjustment.NO_ADJUSTMENT)
        self.assertEqual(schedule.frequency, ScheduleFrequency.MONTHLY)
        self.assertEqual(schedule.end_type, ScheduleEndType.END_DATE)
        self.assertEqual(schedule.interval, 1)
        self.assertEqual(schedule.start_date, interest_start)
        self.assertEqual(schedule.end_date, self.end_date)

        self.assertEqual(interest_start, interest_dates[0])
        self.assertEqual(date(2013, 4, 30), interest_dates[1])
        self.assertEqual(date(2013, 5, 31), interest_dates[2])
        self.assertEqual(date(2014, 2, 28), interest_dates[10])
        self.assertEqual(self.end_date, interest_dates[-1])
        self.assertEqual(schedule.is_due(date(2013, 12, 31)), False)

    def test_daily_schedule_with_end_date(self):
        accrual_schedule = self.__get_daily_schedule(self.start_date, self.end_date, ScheduleEndType.END_DATE)

        accrual_dates = accrual_schedule.get_all_dates(self.end_date);

        self.assertEqual(self.start_date, accrual_dates[0])
        self.assertEqual(self.end_date, accrual_dates[len(accrual_dates) - 1])
        self.assertEqual((self.end_date - self.start_date).days + 1, len(accrual_dates))

        self.assertEqual(accrual_schedule.is_due(self.start_date), True)
        self.assertEqual(accrual_schedule.is_due(self.end_date), True)

        self.assertEqual(accrual_schedule.is_due(self.start_date + relativedelta(days=-1)), False)
        self.assertEqual(accrual_schedule.is_due(self.end_date + relativedelta(days=+1)), False)

    def test_monthly_repeats_schedule(self):
        start_date = date(2019, 12, 1)
        discount_schedule = Schedule(start_date=start_date, end_type=ScheduleEndType.END_REPEATS,
                                     frequency=ScheduleFrequency.MONTHLY, interval=1,
                                     adjustment=BusinessDayAdjustment.NO_ADJUSTMENT, number_of_repeats=3)

        discount_dates = discount_schedule.get_all_dates(self.end_date)

        self.assertEqual(len(discount_dates), 3)
        self.assertEqual(date(2020, 2, 1), discount_dates[2])


if __name__ == '__main__':
    unittest.main()
